@@ -716,6 +716,72 @@ def info_item_uncommmitted(sbox):
     sbox.ospath('newfile'), sbox.ospath('newdir'))
 
 
+def info_item_size_wc_recursive(sbox):
+  "recursive '--show-item=repos-size' on local path"
+
+  sbox.build(read_only=True)
+
+  svntest.actions.run_and_verify_svn(
+    [], [],
+    'info', '--show-item=repos-size', '--recursive',
+    sbox.ospath(''))
+
+
+def info_item_size_repos(sbox):
+  "non-recursive '--show-item=repos-size' on URL"
+
+  sbox.build(read_only=True)
+
+  svntest.actions.run_and_verify_svn(
+    "25\n", [],
+    'info', '--show-item=repos-size',
+    sbox.repo_url + "/iota")
+
+  # Same, but without the newline.
+  svntest.actions.run_and_verify_svn(
+    "25", [],
+    'info', '--show-item=repos-size', '--no-newline',
+    sbox.repo_url + "/iota")
+
+  # Same, but with "human-readable" output.
+  svntest.actions.run_and_verify_svn(
+    "25 B", [],
+    'info', '--show-item=repos-size', '--human-readable',
+    sbox.repo_url + "/iota")
+
+  # No output when the URL is a directory.
+  svntest.actions.run_and_verify_svn(
+    [], [],
+    'info', '--show-item=repos-size',
+    sbox.repo_url)
+
+
+def info_item_size_repos_recursive(sbox):
+  "recursive '--show-item=repos-size' on dir URL"
+
+  sbox.build(read_only=True)
+
+  expected_output = svntest.verify.UnorderedOutput([
+    "25         " + sbox.repo_url + "/iota\n",
+    "27         " + sbox.repo_url + "/A/B/lambda\n",
+    "25         " + sbox.repo_url + "/A/B/E/beta\n",
+    "26         " + sbox.repo_url + "/A/B/E/alpha\n",
+    "23         " + sbox.repo_url + "/A/mu\n",
+    "26         " + sbox.repo_url + "/A/D/gamma\n",
+    "23         " + sbox.repo_url + "/A/D/G/pi\n",
+    "24         " + sbox.repo_url + "/A/D/G/rho\n",
+    "24         " + sbox.repo_url + "/A/D/G/tau\n",
+    "26         " + sbox.repo_url + "/A/D/H/omega\n",
+    "24         " + sbox.repo_url + "/A/D/H/psi\n",
+    "24         " + sbox.repo_url + "/A/D/H/chi\n",
+  ])
+
+  svntest.actions.run_and_verify_svn(
+    expected_output, [],
+    'info', '--show-item=repos-size', '--recursive',
+    sbox.repo_url)
+
+
 def info_item_failures(sbox):
   "failure modes of 'svn info --show-item'"
 
@@ -746,128 +812,11 @@ def info_item_failures(sbox):
     'info', '--show-item=revision', '--no-newline',
     sbox.ospath('A'), sbox.ospath('iota'))
 
+  svntest.actions.run_and_verify_svn(
+    None, (r".*E200007: can't show in-repository size.*"),
+    'info', '--show-item=repos-size',
+    sbox.ospath('iota'))
 
-@Issue(4869)
-def info_tree_conflict_source(sbox):
-  "info --xml: verify source-left, source-right"
-
-  sbox.build()
-  wc_dir = sbox.wc_dir
-
-  B_path = os.path.join(wc_dir, 'A', 'B')
-  lambda_path = os.path.join(B_path, 'lambda')
-  alpha_path = os.path.join(B_path, 'E', 'alpha')
-  F_path = os.path.join(B_path, 'F')
-
-  B2_url = sbox.repo_url + '/A/B2'
-  B2_path = os.path.join(wc_dir, 'A', 'B2')
-  lambda2_path = os.path.join(B2_path, 'lambda')
-
-  # Rev 2 copy B to B2
-  sbox.simple_repo_copy('A/B', 'A/B2')
-  sbox.simple_update()
-
-  # Rev 3:
-  # edit A/B/lambda to test text conflict case
-  # add property to A/B/E/alpha to test property conflict case
-  # rename A/B/F to A/B/Z to test tree conflict case
-
-  svntest.main.file_write(lambda_path, 'B/lambda side of conflict')
-  sbox.simple_propset('blue', 'azul', 'A/B/E/alpha')
-  sbox.simple_move('A/B/F', 'A/B/Z');
-  sbox.simple_commit()
-
-  # Rev 4:
-  # edit A/B2/lambda
-  # add property to A/B2/E/alpha
-  # rename A/B2/F to A/B2/Y
-
-  svntest.main.file_write(lambda2_path, 'B2/lambda side of conflict')
-  sbox.simple_propset('blue', 'bleue', 'A/B2/E/alpha')
-  sbox.simple_move('A/B2/F', 'A/B2/Y');
-  sbox.simple_commit()
-
-  # Now merge B2 into B to cause a text conflict, property conflict, and
-  # tree conflict
-  sbox.simple_update()
-
-  svntest.actions.run_and_verify_svn2(svntest.verify.AnyOutput, [],
-                                      0, 'merge', B2_url, B_path)
-
-  # Verify 'svn info --xml' on the text conflicted case
-
-  exit_code, output, error = svntest.actions.run_and_verify_svn(None,
-                                                                [], 'info',
-                                                                lambda_path,
-                                                                '--xml')
-
-  verify_xml_elements(output,
-                      [('version', {'revision'     : '1',
-                                    'side'         : 'source-left',
-                                    'kind'         : 'file',
-                                    'path-in-repos': 'A/B/lambda',
-                                    'repos-url'    : sbox.repo_url,
-                                   },
-                       )])
-
-  verify_xml_elements(output,
-                      [('version', {'revision'     : '4',
-                                    'side'         : 'source-right',
-                                    'kind'         : 'file',
-                                    'path-in-repos': 'A/B2/lambda',
-                                    'repos-url'    : sbox.repo_url,
-                                   },
-                       )])
-
-  # Verify 'svn info --xml' on the property conflicted case
-
-  exit_code, output, error = svntest.actions.run_and_verify_svn(None,
-                                                                [], 'info',
-                                                                alpha_path,
-                                                                '--xml')
-
-  verify_xml_elements(output,
-                      [('version', {'revision'     : '1',
-                                    'side'         : 'source-left',
-                                    'kind'         : 'file',
-                                    'path-in-repos': 'A/B/E/alpha',
-                                    'repos-url'    : sbox.repo_url,
-                                   },
-                       )])
-
-  verify_xml_elements(output,
-                      [('version', {'revision'     : '4',
-                                    'side'         : 'source-right',
-                                    'kind'         : 'file',
-                                    'path-in-repos': 'A/B2/E/alpha',
-                                    'repos-url'    : sbox.repo_url,
-                                   },
-                       )])
-
-  # Verify 'svn info --xml' on the tree conflicted case
-
-  exit_code, output, error = svntest.actions.run_and_verify_svn(None,
-                                                                [], 'info',
-                                                                F_path,
-                                                                '--xml')
-
-  verify_xml_elements(output,
-                      [('version', {'revision'     : '1',
-                                    'side'         : 'source-left',
-                                    'kind'         : 'dir',
-                                    'path-in-repos': 'A/B/F',
-                                    'repos-url'    : sbox.repo_url,
-                                   },
-                       )])
-
-  verify_xml_elements(output,
-                      [('version', {'revision'     : '4',
-                                    'side'         : 'source-right',
-                                    'kind'         : 'none',
-                                    'path-in-repos': 'A/B2/F',
-                                    'repos-url'    : sbox.repo_url,
-                                   },
-                       )])
 
 ########################################################################
 # Run the tests
@@ -889,8 +838,10 @@ test_list = [ None,
               info_item_simple_multiple,
               info_item_url,
               info_item_uncommmitted,
+              info_item_size_wc_recursive,
+              info_item_size_repos,
+              info_item_size_repos_recursive,
               info_item_failures,
-              info_tree_conflict_source,
              ]
 
 if __name__ == '__main__':
